/* =========================================================================
 *                                randlfib.CPP
 * -------------------------------------------------------------------------
 *  Implementation file for Lagged Fibonacci random number generator as an
 *  object.
 *
 *  Version 0.9
 *
 *  Copyright 1996-1997 by G. Wade Johnson
 *   Use of this code is placed in the public domain as long as the
 *   copyright notice is retained.
 */

#include "environ.h"
#include "randlfib.h"


class  LaggedFibonacciToken : public RandGenTokenImpl
 {
private:
  unsigned          myUpperTap;
  unsigned          myLowerTap;
  unsigned          myLow;
  unsigned          myHigh;
  RingBuffer        myBuffer;
  RandGenToken      myInitializer;

private:
  // protect the = operator and copy constructor so they cannot be used.
  LaggedFibonacciToken &operator=( const LaggedFibonacciToken & /* rhs */ );
  LaggedFibonacciToken( LaggedFibonacciToken const & );

protected:
  friend class LaggedFibonacciRandGen;
  LaggedFibonacciToken( unsigned upperTap,       unsigned lowerTap,
                        unsigned low,            unsigned high,
                        const RingBuffer &Buff,  const RandGenToken  &init
                      );

public:
  virtual   ~LaggedFibonacciToken( void );

  virtual _STRING  CreateInitString( void ) const;
  virtual int   Initialize( istream &InitStream );
  virtual DIFFRET(RandGenTokenImpl,LaggedFibonacciToken)   *Clone() const
   {
    return new LaggedFibonacciToken( myUpperTap, myLowerTap,
                                     myLow,      myHigh,
                                     myBuffer,   myInitializer );

   }

  virtual void  Print( ostream &os ) const;
  virtual void  Read( istream &is );
 } ;



LaggedFibonacciToken::LaggedFibonacciToken(
                        unsigned upperTap,       unsigned lowerTap,
                        unsigned low,            unsigned high,
                        const RingBuffer &Buff,  const RandGenToken  &init
                      )
  : myUpperTap(upperTap), myLowerTap(lowerTap), myLow(low),
    myHigh(high), myBuffer(Buff), myInitializer(init)
 { }




LaggedFibonacciToken::~LaggedFibonacciToken( void )
 { }




_STRING LaggedFibonacciToken::CreateInitString( void ) const
 {
  return myInitializer.CreateInitString();
 }



int   LaggedFibonacciToken::Initialize( istream &InitStream )
 {
  if(!myInitializer.Initialize( InitStream ))
    return 0;

  myHigh = 0;
  myLow  = myUpperTap - myLowerTap;
  myBuffer = RingBuffer( 0 );

  return 1;
 }



void  LaggedFibonacciToken::Print( ostream &os ) const
 {
  os << '(' << myUpperTap << ',' << myLowerTap << ',';
  os << myLow << ',' << myHigh << ',';

  if(!myBuffer.max_size())
   {
    os << "*,";
   }
  else
   {
    myBuffer.Print( os );
   }
  os << ',';

  myInitializer.Print( os );
  os << ')';
 }



void  LaggedFibonacciToken::Read( istream &is )
 {
  char   Dummy;
  is >> Dummy >> myUpperTap >> Dummy >> myLowerTap;
  is >> Dummy >> myLow >> Dummy >> myHigh >> Dummy;

  is >> Dummy;
  if(Dummy == '*')
   {
    myBuffer = RingBuffer( 0 );
   }
  else
   {
    myBuffer.Read( is );
   }
  is >> Dummy;

  myInitializer.Read( is );
  is >> Dummy;
 }





LaggedFibonacciRandGen::LaggedFibonacciRandGen( unsigned lower, unsigned upper,
                                                       RandomGenerator *pRng )
 : myUpperTap(upper), myLowerTap(lower), myBuffer(upper),
   myHigh(myBuffer.MakeIterator(0)), myLow(myBuffer.MakeIterator(upper-lower)),
   myInitializer(pRng)
 {
  //  Here's where the fun begins
  LaggedFibonacciRandGen::Reset();
 }



LaggedFibonacciRandGen::~LaggedFibonacciRandGen( void )
 {
  //  Clean up after yourself.
  delete myInitializer;
  myInitializer = 0;
 }


//  This is where the rest of the functionality is defined.


void     LaggedFibonacciRandGen::IncrTaps()
 {
  ++myLow;
  ++myHigh;
 }





void     LaggedFibonacciRandGen::InitArray( void )
 {
  myBuffer.Init();

  for(unsigned i=0;i<myUpperTap;i++)
    myBuffer.add( myInitializer->Number() );
 }


RandGenToken  LaggedFibonacciRandGen::CreateToken( void ) const
 {
  return RandGenToken(
         new LaggedFibonacciToken( myUpperTap, myLowerTap, myLow.Index(),
                                   myHigh.Index(), myBuffer,
                                   myInitializer->CreateToken()
                                 )
                     );
 }



int     LaggedFibonacciRandGen::SetToken( const RandGenToken &Token )
 {
  const LaggedFibonacciToken *pTok =
           DYNACAST(const LaggedFibonacciToken *)(Token.Implementation());

  if(pTok)
   {
    if(!myInitializer->SetToken( pTok->myInitializer ))
      return 0;

    if(pTok->myBuffer.max_size())
     {
      myBuffer = pTok->myBuffer;

      myLow  = myBuffer.MakeIterator( pTok->myLow  );
      myHigh = myBuffer.MakeIterator( pTok->myHigh );
     }
    else
     {
      InitArray();
      InitTaps();
     }

    return 1;
   }

  return 0;
 }



void    LaggedFibonacciRandGen::Reset( void )
 {
  myInitializer->Reset();

  InitArray();

  InitTaps();
 }



void    LaggedFibonacciRandGen::IntSeed( size_rand seed )
 {
  myInitializer->IntSeed( seed );

  InitArray();
  InitTaps();
 }



size_rand   LaggedFibonacciRandGen::Number( void )
 {
  size_rand  RandNum = LowValue() + HighValue();

  IncrTaps();
  SetNewElement( RandNum );

  return RandNum;
 }
