/* =========================================================================
 *                                 RANDSHA.CPP
 * -------------------------------------------------------------------------
 *  Implementation file for an abstract congruential pseudo-random number
 *  generator with a modulus of 2^32 as an object.
 *
 *  Version 0.9
 *
 *  Copyright 1995-1997 by G. Wade Johnson
 *   Use of this code is placed in the public domain as long as the
 *   copyright notice is retained.
 */

#include <string.h>

#include "environ.h"

#include _STRSTREAM_HEADER

#include "randsha.h"


class  SHARandGenToken : public RandGenTokenImpl
 {
public:
  virtual _STRING  CreateInitString( void ) const;
  virtual int   Initialize( istream &InitStream );
  virtual DIFFRET(RandGenTokenImpl,SHARandGenToken)   *Clone() const
   {
    return new SHARandGenToken( myBufferSize, mySeedBuffer );
   }

  void       Print( ostream &os ) const
   {
    os << '(' << myBufferSize << ',';
    os.write( mySeedBuffer, myBufferSize );
    os << ')' << flush;
   }
  void       Read( istream &is )
   {
    char  dummy;
    delete[] mySeedBuffer;

    is >> dummy >> myBufferSize >> dummy;
    mySeedBuffer = new char[myBufferSize];
    is.read( mySeedBuffer, myBufferSize );
    is >> dummy;
   } 

public:
  SHARandGenToken( unsigned short size, const char *seedBuff ) :
    myBufferSize(size), mySeedBuffer(0)
   {
    mySeedBuffer = new char[size];
    memcpy( mySeedBuffer, seedBuff, size );
   }
  virtual ~SHARandGenToken( void ) { delete[] mySeedBuffer; }

protected:
  unsigned short  BufferSize() const { return myBufferSize; }
  char           *SeedBuffer() const { return mySeedBuffer; }

private:
  friend class SHARandGen;

  unsigned short  myBufferSize;
  char           *mySeedBuffer;
 } ;





_STRING   SHARandGenToken::CreateInitString( void ) const
 {
  ostrstream  aString;

  aString << '(' << myBufferSize << ',';
  aString.write( mySeedBuffer, myBufferSize );
  aString << ')' << ends;

  return _STRING( aString.str() );
 }




int   SHARandGenToken::Initialize( istream &InitStream )
 {
  Read( InitStream );

  return 1;
 }




void         SHARandGen::SetSeedBuffer( unsigned short size,
                                         const char *pBuffer )
 {
  mySeedBuffer = new char[size+sizeof(size_rand)+1];
  myBufferSize = size;

  memcpy( mySeedBuffer, pBuffer, size );
  pNumber = (size_rand *)(mySeedBuffer+size);
  mySeedBuffer[size+sizeof(size_rand)] = 0;
 }


RandGenToken    SHARandGen::CreateToken( void ) const
 {
  return RandGenToken( new SHARandGenToken( myBufferSize, mySeedBuffer ) );
 }



int   SHARandGen::SetToken( const RandGenToken &tok )
 {
  const SHARandGenToken  *pToken =
                   DYNACAST( const SHARandGenToken * )(tok.Implementation());

  if(pToken)
   {
    SetSeedBuffer( pToken->myBufferSize, pToken->mySeedBuffer );
    return 1;
   }

  return 0;
 }




size_rand   SHARandGen::Number( void )
 {
  *pNumber = NextNumber();

  shsInit( &mySHS );
  shsUpdate( &mySHS, (unsigned char *)mySeedBuffer,
             myBufferSize+sizeof(size_rand) );
  shsFinal( &mySHS );

  return (*(size_rand *)(mySHS.digest));
 }





class  SHACtrRandGenToken : public SHARandGenToken
 {
public:
  virtual _STRING  CreateInitString( void ) const;
  virtual int   Initialize( istream &InitStream );
  virtual DIFFRET(RandGenTokenImpl,SHACtrRandGenToken)   *Clone() const
   {
    return new SHACtrRandGenToken( BufferSize(), SeedBuffer(), mySeed,
                                   myCurrSeed );
   }

  void       Print( ostream &os ) const
   {
    os << '(';
    SHARandGenToken::Print( os );
    os << ',' << mySeed << ',' << myCurrSeed << ')' << flush;
   }
  void       Read( istream &is )
   {
    char  dummy;

    is >> dummy;
    SHARandGenToken::Read( is );
    is >> dummy >> mySeed >> dummy >> myCurrSeed >> dummy;
   } 

public:
  SHACtrRandGenToken( unsigned short size, const char *seedBuff, size_rand seed,
                       size_rand currSeed ) :
    SHARandGenToken(size,seedBuff), mySeed(seed), myCurrSeed(currSeed)
   {
   }
  virtual ~SHACtrRandGenToken( void ) { }

private:
  friend class SHACtrRandGen;

  size_rand   mySeed;
  size_rand   myCurrSeed;
 } ;


_STRING   SHACtrRandGenToken::CreateInitString( void ) const
 {
  ostrstream  aString;

  aString << '(' << SHARandGenToken::CreateInitString() << ',';
  aString << mySeed << ',' << mySeed << ')' << ends;

  return _STRING( aString.str() );
 }




int   SHACtrRandGenToken::Initialize( istream &InitStream )
 {
  Read( InitStream );

  return 1;
 }




DIFFRET(RandomGenerator,SHACtrRandGen)  *SHACtrRandGen::ReCreate( void ) const
 {
  return new SHACtrRandGen( BufferSize(), SeedBuffer(), mySeed );
 }





RandGenToken    SHACtrRandGen::CreateToken( void ) const
 {
  return RandGenToken( new SHACtrRandGenToken( BufferSize(), SeedBuffer(),
                                  mySeed, myCurrSeed )
                     );
 }



int   SHACtrRandGen::SetToken( const RandGenToken &tok )
 {
  const SHACtrRandGenToken  *pToken =
             DYNACAST( const SHACtrRandGenToken * )(tok.Implementation());

  if(pToken)
   {
    if(!SHARandGen::SetToken( tok ))
      return 0;

    SetSeed( pToken->mySeed );
    SetCurrSeed( pToken->myCurrSeed );
    return 1;
   }

  return 0;
 }





class  SHARngRandGenToken : public SHARandGenToken
 {
public:
  virtual _STRING   CreateInitString( void ) const;
  virtual int   Initialize( istream &InitStream );
  virtual DIFFRET(RandGenTokenImpl,SHARngRandGenToken)   *Clone() const
   {
    return new SHARngRandGenToken( BufferSize(), SeedBuffer(), myToken );
   }

  void       Print( ostream &os ) const
   {
    os << '(';
    SHARandGenToken::Print( os );
    os << ',' << myToken << ')' << flush;
   }
  void       Read( istream &is )
   {
    char  dummy;

    is >> dummy;
    SHARandGenToken::Read( is );
    is >> dummy >> myToken;
   }

public:
  SHARngRandGenToken( unsigned short size, const char *seedBuff,
                                                const RandGenToken &Token )
    : SHARandGenToken(size,seedBuff), myToken(Token)
   {
   }
  virtual ~SHARngRandGenToken( void ) { }

private:
  friend class SHARngRandGen;

  RandGenToken  myToken;
 } ;






_STRING   SHARngRandGenToken::CreateInitString( void ) const
 {
  ostrstream  aString;

  aString << '(' << SHARandGenToken::CreateInitString() << ',';
  aString << myToken.CreateInitString() << ')' << ends;

  return _STRING( aString.str() );
 }




int   SHARngRandGenToken::Initialize( istream &InitStream )
 {
  Read( InitStream );

  return 1;
 }








DIFFRET(RandomGenerator,SHARngRandGen)  *SHARngRandGen::ReCreate( void ) const
 {
  return new SHARngRandGen( BufferSize(), SeedBuffer(), myRNG );
 }





RandGenToken    SHARngRandGen::CreateToken( void ) const
 {
  return RandGenToken( new SHARngRandGenToken( BufferSize(), SeedBuffer(),
                                  myRNG->CreateToken() )
                     );
 }



int   SHARngRandGen::SetToken( const RandGenToken &tok )
 {
  const SHARngRandGenToken  *pToken =
              DYNACAST( const SHARngRandGenToken * )(tok.Implementation());

  if(pToken)
   {
    if(SHARandGen::SetToken( tok ) && myRNG->SetToken( pToken->myToken ))
      return 1;
   }

  return 0;
 }
