/* =========================================================================
 *                                 RANDSHA.H
 * -------------------------------------------------------------------------
 *  Interface file for a pseudo-random number generator based on the SHA
 *  algorithm as an object.
 *
 *  Version 0.9
 *
 *  Copyright 1995-1997 by G. Wade Johnson
 *   Use of this code is placed in the public domain as long as the
 *   copyright notice is retained.
 */

#if !defined( RANDSHA_H )
  #define RANDSHA_H

  #include <string.h>
  #if !defined( RANDGEN_H )
    #include "randgen.h"
  #endif
  extern "C"
   {
    #include "sha.h"
   }


  class  SHARandGen : public RandomGenerator
   {
  private:
    unsigned short        myBufferSize;
    char                 *mySeedBuffer;
    size_rand            *pNumber;
    SHS_INFO              mySHS;

  private:
    // protect the = operator and copy constructor so they cannot be used.
    SHARandGen& operator=( SHARandGen const & );
    SHARandGen( SHARandGen const & );

  protected:
    unsigned short        BufferSize( void ) const { return myBufferSize; }
    const char           *SeedBuffer( void ) const { return mySeedBuffer; }

    void         SetSeedBuffer( unsigned short size, const char *pBuffer );
    void         SetSeedBuffer( const char *pszString )
     {
      SetSeedBuffer( strlen( pszString ), pszString );
     }
    virtual size_rand  NextNumber( void ) = 0;

  public:
    EXPLICIT SHARandGen( const char *seedStr ) : mySeedBuffer(0)
     {
      myBufferSize = strlen( seedStr );
      SetSeedBuffer( seedStr );
      SHARandGen::Reset();
     }
    SHARandGen( unsigned short size, const char *seedStr ) :
     myBufferSize(size), mySeedBuffer(0)
     {
      SetSeedBuffer( size, seedStr );
      SHARandGen::Reset();
     }
    virtual   ~SHARandGen() { delete[] mySeedBuffer; }

    virtual RandGenToken  CreateToken( void ) const;
    virtual int     SetToken( const RandGenToken &Token );

    virtual  size_rand   Number( void );
   } ;


  class  SHACtrRandGen : public SHARandGen
   {
  private:
    size_rand      mySeed;
    size_rand      myCurrSeed;

  private:
    // protect the = operator and copy constructor so they cannot be used.
    SHACtrRandGen& operator=( SHACtrRandGen const & );
    SHACtrRandGen( SHACtrRandGen const & );

  protected:
    size_rand    Seed( void )      const       { return mySeed; }
    size_rand    CurrSeed( void )  const       { return myCurrSeed; }
    void         SetSeed( size_rand seed )     { mySeed = seed; }
    void         SetCurrSeed( size_rand seed ) { myCurrSeed = seed; }

    virtual size_rand  NextNumber( void )  { return myCurrSeed++; }

  public:
    EXPLICIT SHACtrRandGen( const char *seedStr="SHAPRNG", size_rand seed=0 )
     : SHARandGen(seedStr), mySeed(seed)
     {
      SHACtrRandGen::Reset();
     }
    SHACtrRandGen( unsigned short size, const char *seedStr, size_rand seed=0 )
     : SHARandGen(size,seedStr), mySeed(seed)
     {
      SHACtrRandGen::Reset();
     }
    virtual   ~SHACtrRandGen() {  }
 
    virtual DIFFRET(RandomGenerator,SHACtrRandGen)    *ReCreate( void ) const;

    virtual RandGenToken  CreateToken( void ) const;
    virtual int     SetToken( const RandGenToken &Token );

    virtual void    Reset( void ) { myCurrSeed = mySeed; }

    virtual  void   IntSeed( size_rand seed ) { myCurrSeed = seed; }
   } ;



  class  SHARngRandGen : public SHARandGen
   {
  private:
    RandomGenerator  *myRNG;

  private:
    // protect the = operator and copy constructor so they cannot be used.
    SHARngRandGen& operator=( SHARngRandGen const & );
    SHARngRandGen( SHARngRandGen const & );

  protected:
    RandomGenerator   *pRNG( void )  const  { return myRNG; }

    virtual size_rand  NextNumber( void )  { return myRNG->Number(); }

  public:
    SHARngRandGen( const char *seedStr, RandomGenerator *pRng )
     : SHARandGen(seedStr), myRNG(pRng)
     {
      SHARngRandGen::Reset();
     }
    SHARngRandGen( unsigned short size, const char *seedStr,
                                         RandomGenerator *pRng )
     : SHARandGen(size,seedStr), myRNG(pRng)
     {
      SHARngRandGen::Reset();
     }
    virtual ~SHARngRandGen() { delete myRNG; }
 
    virtual DIFFRET(RandomGenerator,SHARngRandGen)    *ReCreate( void ) const;

    virtual RandGenToken  CreateToken( void ) const;
    virtual int     SetToken( const RandGenToken &Token );

    virtual void    Reset( void ) { myRNG->Reset(); }

    virtual void   IntSeed( size_rand seed ) { myRNG->IntSeed( seed ); }
   } ;



#endif
