/* =========================================================================
 *                             RANDSHUF.CPP
 * -------------------------------------------------------------------------
 *  Implementation file for the Shuffled pseudo-random number generator
 *  object.  (Based on Algorithm described in Knuth.)
 *
 *  Version 0.9
 *
 *  Copyright 1995-1997 by G. Wade Johnson
 *   Use of this code is placed in the public domain as long as the
 *   copyright notice is retained.
 */

#include <string.h>
#include <memory.h>

#include "environ.h"

#include _STRSTREAM_HEADER

#include "randshuf.h"




class  ShuffledMRandGenToken : public RandGenTokenImpl
 {
public:
  virtual   ~ShuffledMRandGenToken( void )
   {
    delete[] pmyArray;
    pmyArray = 0;
   }

  virtual _STRING   CreateInitString( void ) const;
  virtual int   Initialize( istream &InitStream );
  virtual DIFFRET(RandGenTokenImpl,ShuffledMRandGenToken)   *Clone() const
   {
    return new ShuffledMRandGenToken( mySize, pmyArray, myTokenCombine );
   }

  void       Print( ostream &os ) const;
  void       Read( istream &is );

private:
  // Don't define. These functions are not to be allowed.
  ShuffledMRandGenToken( const ShuffledMRandGenToken &rhs );
  ShuffledMRandGenToken& operator=( const ShuffledMRandGenToken &rhs );

  friend class ShuffledMRandGen;
  ShuffledMRandGenToken( unsigned size, size_rand *pArray,
                                                  const RandGenToken &Tok ) :
    mySize(size), myTokenCombine(Tok)
   {
    pmyArray = new size_rand[mySize];
    if(!pmyArray)
      return;

    memcpy( pmyArray, pArray, mySize*sizeof(size_rand) );
   }

  unsigned        mySize;
  size_rand      *pmyArray;
  RandGenToken    myTokenCombine;
 } ;



void   ShuffledMRandGenToken::Print( ostream &os ) const
 {
  os << '(' << mySize;
  if(mySize)
   {
    char       c = '(';
    unsigned   i;
    for(i=0;i<mySize;i++)
     {
      os << c << pmyArray[i];
      c = ',';
     }
    os << ')';
   }
  os << ',' << myTokenCombine << ')' << flush;
 }




void   ShuffledMRandGenToken::Read( istream &is )
 {
  char  dummy;
  is >> dummy >> mySize;
  delete[] pmyArray;
  pmyArray = 0;

  if(mySize)
   {
    unsigned   i;

    pmyArray = new size_rand[mySize];
    if(!pmyArray)
      return;

    for(i=0;i<mySize;i++)
     {
      is >> dummy >> pmyArray[i];
     }
    is >> dummy;
   }
  is >> dummy >> myTokenCombine >> dummy;
 }




_STRING   ShuffledMRandGenToken::CreateInitString( void ) const
 {
  ostrstream  String;

  String << '(' << mySize << ',' << myTokenCombine.CreateInitString()
         << ')' << ends;

  return _STRING( String.str() );
 }






int   ShuffledMRandGenToken::Initialize( istream &InitStream )
 {
  char    junk;

  InitStream >> ws >> junk >> ws >> mySize >> ws >> junk;

  if(!myTokenCombine.Initialize( InitStream ))
    return 0;

  InitStream >> ws >> junk;

  delete[] pmyArray;
  pmyArray = 0;

  return 1;
 }



DIFFRET(RandomGenerator,ShuffledMRandGen)
                                     *ShuffledMRandGen::ReCreate( void ) const
 {
  return new ShuffledMRandGen( mySize, pRandGenX()->ReCreate(),
                                      pRandGenY()->ReCreate() );
 }




ShuffledMRandGen::ShuffledMRandGen( unsigned size, RandomGenerator *pGenX,
                                                 RandomGenerator *pGenY ) :
  CombineRandGen(pGenX,pGenY), mySize(size), myBadFlag(false)
 {
  pRandArray = new size_rand[size];
  if(!pRandArray)
   {
    MakeBad();
    return;
   }

  ShuffledMRandGen::Reset();
 }




ShuffledMRandGen::~ShuffledMRandGen()
 {
  delete[] pRandArray;
 }



RandGenToken  ShuffledMRandGen::CreateToken( void ) const
 {
  return RandGenToken(
         new ShuffledMRandGenToken( mySize, pRandArray,
                                   CombineRandGen::CreateToken()
                                  )
                     );
 }



int     ShuffledMRandGen::SetToken( const RandGenToken &Token )
 {
  const ShuffledMRandGenToken *pSRGTok =
             DYNACAST( const ShuffledMRandGenToken * )(Token.Implementation());

  if(pSRGTok && mySize == pSRGTok->mySize)
   {
    if(!CombineRandGen::SetToken( pSRGTok->myTokenCombine ))
      return 0;

    if(pSRGTok->pmyArray)
      memcpy( pRandArray, pSRGTok->pmyArray, mySize*sizeof(size_rand) );
    else
      Reset();

    return 1;
   }

  return 0;
 }


void        ShuffledMRandGen::Reset( void )
 {
  unsigned  index;

  CombineRandGen::Reset();

  for(index=0;index < mySize;index++)
   {
    pRandArray[index] = pRandGenX()->Number();
   }
 }




size_rand   ShuffledMRandGen::Number( void )
 {
  size_rand  Y;
  unsigned   j = (unsigned)pRandGenY()->LimitedNumber( (unsigned long)mySize );

  Y = pRandArray[j];
  pRandArray[j] = pRandGenX()->Number();

  return Y;
 }
