/* =========================================================================
 *                                 RANDTABL.CPP
 * -------------------------------------------------------------------------
 *  Implementation file for any table driven pseudo-random number generator
 *  as an object.  (Assumes table on disk.)
 *
 *  Version 0.9
 *
 *  Copyright 1995-1997 by G. Wade Johnson
 *   Use of this code is placed in the public domain as long as the
 *   copyright notice is retained.
 */

#include <ctype.h>

#include "environ.h"
#include _FSTREAM_HEADER
#include _STRSTREAM_HEADER

#include <string.h>

#include "randtabl.h"




const char  szTableMarker[] = "Begin Table";

const int  LineSize = 128;
const int  MaxFileName = 255;




class  TableRandGenToken : public RandGenTokenImpl
 {
public:
  virtual   ~TableRandGenToken( void ) { delete[] myFileName; }

  virtual _STRING   CreateInitString( void ) const;
  virtual int   Initialize( istream &InitStream );
  virtual DIFFRET(RandGenTokenImpl,TableRandGenToken)   *Clone() const
   {
    return new TableRandGenToken( myFileName, myStart, myPosition );
   }

  void       Print( ostream &os ) const
   {
    os << '(' << myFileName << ',' << (long)myStart
       << ',' << (long)myPosition << ')';
   }
  void       Read( istream &is )
   {
    char  dummy;
    char  filename[MaxFileName];
    unsigned long   temp;

    is >> dummy;
    is.get( filename, MaxFileName, ',' );
    is >> dummy >> temp >> dummy;
    myStart = (streampos)temp;
    is >> temp >> dummy;
    myPosition = (streampos)temp;

    delete[] myFileName;
    myFileName = new char[strlen(filename)+1];
    strcpy( myFileName, filename );
   } 

private:
  friend class TableRandGen;
  TableRandGenToken( const char *pszFileName, streampos start,
                                                        streampos position ) :
    myPosition( position ), myStart( start )
   {
    myFileName = new char[strlen(pszFileName)+1];
    strcpy( myFileName, pszFileName );
   }

  char           *myFileName;
  streampos       myPosition;
  streampos       myStart;
 } ;





_STRING   TableRandGenToken::CreateInitString( void ) const
 {
  ostrstream  String;

  String << myFileName << ',' << (long)myStart << ends;

  return _STRING( String.str() );
 }




int   TableRandGenToken::Initialize( istream &InitStream )
 {
  char    dummy;
  char    filename[MaxFileName];
  long    temp;

  InitStream >> ws;
  InitStream.get( filename, MaxFileName, ',' );
  InitStream >> ws >> dummy >> temp;
  myStart = (streampos)temp;
  myPosition = (streampos)0;

  delete[] myFileName;
  myFileName = new char[strlen(filename)+1];
  strcpy( myFileName, filename );

  return 1;
 }






TableRandGen::TableRandGen( const char *szFileName, streampos start ) :
  myTable( szFileName ), myStart(start), myBadFlag(false)
 {
  if(!myTable)
   {
    MakeBad();
    return;
   }

  myFileName = new char[strlen(szFileName)+1];
  strcpy( myFileName, szFileName );

  TableRandGen::Reset();

  return;
 }





void        TableRandGen::Reset( void )
 {
  myTable.seekg( myStart, ios::cur );

  return;
 }





RandGenToken    TableRandGen::CreateToken( void ) const
 {
  return  RandGenToken(  new TableRandGenToken( myFileName, myStart,
                          (CONSTCAST(TableRandGen*)(this))->myTable.tellg() )
                      );
 }



int   TableRandGen::SetToken( const RandGenToken &tok )
 {
  const TableRandGenToken  *pToken =
                 DYNACAST( const TableRandGenToken * )(tok.Implementation());

  if(pToken && strcmp( myFileName, pToken->myFileName ) == 0)
   {
    myStart = pToken->myStart;

    if(pToken->myPosition != (streampos)0)
      myTable.seekg( pToken->myPosition, ios::beg );
    else
      Reset();

    return 1;
   }

  return 0;
 }





void     TableRandGen::IntSeed( size_rand seed )
 {
  myStart = (streampos)seed;
  Reset();

  return;
 }




TextTableRandGen::TextTableRandGen( const char *szFileName, streampos start ) :
  TableRandGen( szFileName, start )
 {
  TextTableRandGen::Reset();

  return;
 }





void        TextTableRandGen::Reset( void )
 {
  // Find the beginning of the random number set.
  InitTable();

  Table().seekg( Start(), ios::cur );

  return;
 }





void     TextTableRandGen::InitTable( void )
 {
  ifstream  &is = Table();
  char   szLine[LineSize+1];

  is.seekg( (streampos)0, ios::beg );

  do
   {
    is.getline( szLine, LineSize );
   } while(!is.eof() && strstr( szLine, szTableMarker ) == NULL);

  return;
 }





size_rand   BinaryTableRandGen::Number( void )
 {
  size_rand  num;

  Table().read( (char*)&num, sizeof(num) );

  if(Table().eof())
    Table().seekg( (streampos)0, ios::beg );

  return num;
 }
