/* ========================================================================
 *                                RINGBUFF.H
 * ------------------------------------------------------------------------
 *  Interface for a ring buffer class with iterators.
 *
 *  Version 0.9
 *
 *  Copyright 1997 by G. Wade Johnson
 *   Use of this code is placed in the public domain as long as the
 *   copyright notice is retained.
 */

#if !defined( RINGBUFF_H )
  #define  RINGBUFF_H

  #if !defined( ENVIRON_H )
    #include "environ.h"
  #endif
  #include <assert.h>

  class  RingBuffer
   {
    unsigned        mySize;
    unsigned long  *myArray;
    unsigned        myCursor;
    BOOL            myFirstPass;

  protected:
    const unsigned long &operator()( unsigned index ) const
     {
      return myArray[index];
     }
    unsigned long &operator()( unsigned index )
     {
      return myArray[index];
     }
    BOOL      First()  const { return myFirstPass; }
    unsigned  Cursor() const { return myCursor; }

  public:
    EXPLICIT RingBuffer( unsigned Size )
     : mySize(Size), myArray(0), myCursor(0), myFirstPass(true)
     {
      if(!Size)
        Size = 1;

      myArray = new unsigned long[Size];
     }
    RingBuffer( const RingBuffer &rhs );
    ~RingBuffer()
     {
      delete[] myArray;
      myArray = 0;
     }

    void      Init()  { myCursor = 0; myFirstPass = true; }
    unsigned  max_size()   const { return mySize; }
    unsigned  size()       const
     {
      return myFirstPass ? myCursor : mySize;
     }
    void      add( unsigned long val )
     {
      myArray[myCursor] = val;
      if(++myCursor >= max_size())
       {
        myCursor = 0;
        myFirstPass = false;
       }
     }
    BOOL      empty() const  { return myFirstPass && 0 == myCursor; }

    RingBuffer &operator=( const RingBuffer &rhs );

    class Iterator
     {
      RingBuffer  *myBuffer;
      unsigned     myIndex;

      Iterator( RingBuffer *mine, unsigned index )
       : myBuffer(mine), myIndex(index)
       {
        index %= myBuffer->max_size();
       }

    public:
      friend RingBuffer;
      Iterator() : myBuffer(0), myIndex((unsigned)-1)
       { }
      Iterator( const Iterator &rhs )
       {
        *this = rhs;
       }
      ~Iterator()
       { }

      unsigned   Index()  const  { return myIndex; }

      const unsigned long &operator *() const
       {
        assert(myBuffer && (unsigned)-1 != myIndex);
        return (*myBuffer)(myIndex);
       }
      unsigned long &operator *()
       {
        assert(myBuffer && (unsigned)-1 != myIndex);
        return (*myBuffer)(myIndex);
       }

      Iterator operator++();
      Iterator operator++( int );
      Iterator operator--();
      Iterator operator--( int );

      Iterator &operator=( const Iterator &rhs );

      BOOL operator==( const Iterator &rhs )
       {
        return ((unsigned)-1 == myIndex && (unsigned)-1 == rhs.myIndex)
              || ((myIndex == rhs.myIndex) && (myBuffer == rhs.myBuffer));
       }
      BOOL operator!=( const Iterator &rhs )
       {
        return !(*this == rhs);
       }
     } ;

    friend Iterator;
    Iterator  MakeIterator( unsigned index )
     {
      return Iterator( this, index%max_size() );
     }
    Iterator  begin();
    Iterator  end()    { return Iterator( this, (unsigned)-1 ); }

    void   Print( ostream &os ) const;
    void   Read( istream &is );
   } ;

  ostream &operator<<( ostream &os, const RingBuffer &r );
  istream &operator>>( istream &is, RingBuffer &r );

#endif
