/* =========================================================================
 *                                 RNDDWYER.H
 * -------------------------------------------------------------------------
 *  Interface file for any linear congruential pseudo-random number generator
 *  as an object.
 *
 *  This generator is based on the algorithm described in Jerry Dwyer's
 *  article in CUJ June 1995 entitled
 *     "Quick and Portable Random Number Generators".
 *
 *  WARNING: only valid for  Modulus > Multiplier * Multiplier.
 *
 *  Version 0.9
 *
 *  Copyright 1995-1997 by G. Wade Johnson
 *   Use of this code is placed in the public domain as long as the
 *   copyright notice is retained.
 */

#if !defined( RNDDWYER_H )
  #define RNDDWYER_H

  #if !defined( RANDGEN_H )
    #include "randgen.h"
  #endif
  #if !defined( RNDMCONM_H )
    #include "rndmconm.h"
  #endif


  class  DwyerRandGen : public MultConModRandGen
   {
  private:
   //  Used in Number() to prevent overflow.
    const unsigned long   myQuotient;
    const unsigned long   myRemainder;

  private:
    // protect the = operator and copy constructor so they cannot be used.
    DwyerRandGen& operator=( DwyerRandGen const & );
    DwyerRandGen( DwyerRandGen const & );

  protected:
    unsigned long   Quotient( void )  const { return myQuotient; }
    unsigned long   Remainder( void ) const { return myRemainder; }

  public:
    DwyerRandGen( unsigned long mult, unsigned long mod, size_rand seed=1 ) :
      MultConModRandGen( mult, mod, seed ),
      myQuotient(mod/mult), myRemainder(mod - myQuotient*mult)
     {
      SetSeed( seed % (DwyerRandGen::MaxRandom() + 1) );
      DwyerRandGen::Reset();
     }
    virtual   ~DwyerRandGen() { }
 
    virtual DIFFRET(RandomGenerator,DwyerRandGen)  *ReCreate( void ) const;

    virtual  void        Reset( void ) { SetCurrSeed( Seed() ); }

    virtual  size_rand   Number( void );
    virtual  size_rand   MaxRandom( void ) const { return Modulus()-1; }
   } ;


/* -------------------------------------------------------------------------
 * The following PRNGs are taken from Jerry Dwyer's article in CUJ June 1995
 *  entitled "Quick and Portable Random Number Generators".
 */

 class Dwyer1RandGen : public DwyerRandGen
  {
 public:
   EXPLICIT Dwyer1RandGen( size_rand seed=1 ) :
    DwyerRandGen( 41358UL, 0x7FFFFFFFUL, seed )
    { }
  } ;



 class Dwyer2RandGen : public DwyerRandGen
  {
 public:
   EXPLICIT Dwyer2RandGen( size_rand seed=1 ) :
    DwyerRandGen( 48271UL, 0x7FFFFFFFUL, seed )
    { }
  } ;



 class Dwyer3RandGen : public DwyerRandGen
  {
 public:
   EXPLICIT Dwyer3RandGen( size_rand seed=1 ) :
    DwyerRandGen( 69621UL, 0x7FFFFFFFUL, seed )
    { }
  } ;



 class Dwyer4RandGen : public DwyerRandGen
  {
 public:
   EXPLICIT Dwyer4RandGen( size_rand seed=1 ) :
    DwyerRandGen( 16807UL, 0x7FFFFFFFUL, seed )
    { }
  } ;



#endif
